! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: dssremove_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: A tasklet to remove the DSS from a block.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module dssremove_mod

  use types

  use param
  use dal
  use errorHandling
  use dss

  use dss_aux	! for dumpDss
  use dal_aux	! for splitSetTabName

  implicit none 

  private 

  ! Command-line parameters:
  character(255) ::&
    setName
  character(20) ::&
    extensionType,&
    tabName,&
    arrayName
  integer ::&
    extensionNum

  public dssremove

contains 

  !---------------------------------------------------------------------
  subroutine dssremove
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT) :: set
    type(BlockT) :: dssBlock
    type(DssT) ::&
      dssPointer

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    set = dataSet(setName, MODIFY)

    select case(extensionType)
    case('bintabname')
      dssBlock = block(set, tabName)!, READ)

    case('arrayname')
      dssBlock = block(set, arrayName)!, READ)

    case('number')
      dssBlock = block(set, extensionNum)!, READ)

    case default
call error('badExtType',"The value of the --exttype parameter is not recognized.")
    end select

    if (.not. hasDss(dssBlock))& ! in ssclib/dss_aux
call error('noDss',"The block (described by table or array name, or extension number) contains no DSS")

    dssPointer = dataSubspace(dssBlock)
    call dssDelete(dssBlock) ! in dsslib

    call release(set)
    call dssRelease(dssPointer)

  end subroutine dssremove

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    setName = stringParameter('set')
    extensionType = stringParameter('exttype')
    select case(extensionType)
    case('bintabname')
      tabName = stringParameter('table')

    case('arrayname')
      arrayName = stringParameter('array')

    case('number')
      extensionNum = intParameter('extn')

    case default
call error('badExtType',"The value of the --exttype parameter is not recognized.")
    end select

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module dssremove_mod
