! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: test_utils
!
! Authors:  Masaaki Sakano - SSC/LUX
!          
! Description: Some useful routines for tests (debugging).
! 
! Comments: TODO:
! 
! $Id: test_utils.f90,v 0.0 1998/05/27 13:45:39 gvacanti Exp $
!- 
module test_utils

  use types	! for int8, single, double, etc.
  use errorhandling

  implicit none 

  private
  public isNearlyEqual	! See function isNearlyEqualDDCore() for description.

  interface isNearlyEqual
    module procedure isNearlyEqualInt8Int8
    module procedure isNearlyEqualInt8Int16, isNearlyEqualInt16Int16
    module procedure isNearlyEqualInt16Int8 
    module procedure isNearlyEqualInt8Int32, isNearlyEqualInt16Int32
    module procedure isNearlyEqualInt32Int32, isNearlyEqualInt32Int16
    module procedure isNearlyEqualInt32Int8
    module procedure isNearlyEqualInt8Single, isNearlyEqualInt16Single
    module procedure isNearlyEqualInt32Single, isNearlyEqualSingleSingle
    module procedure isNearlyEqualSingleInt32, isNearlyEqualSingleInt16
    module procedure isNearlyEqualSingleInt8
    module procedure isNearlyEqualInt8Double, isNearlyEqualInt16Double
    module procedure isNearlyEqualInt32Double, isNearlyEqualSingleDouble
    module procedure isNearlyEqualDoubleDouble, isNearlyEqualDoubleSingle
    module procedure isNearlyEqualDoubleInt32, isNearlyEqualDoubleInt16
    module procedure isNearlyEqualDoubleInt8
  end interface

contains

  !---------------------------------------------------------------------
  logical function isNearlyEqualInt8Int8(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    integer(int8),  intent(in) :: cmp, compared
    integer,        intent(in) :: precision
    integer(int32) :: cmpl

    cmpl = int(cmp, kind=int32)
    isNearlyEqual = isNearlyEqualLLCore(cmpl, int(compared, kind=int32)&
       , cmpl, precision)
  end function isNearlyEqualInt8Int8

  !---------------------------------------------------------------------
  logical function isNearlyEqualInt8Int16(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    integer(int8),  intent(in) :: cmp
    integer(int16), intent(in) ::      compared
    integer,        intent(in) :: precision
    integer(int32) :: cmpl

    cmpl = int(cmp, kind=int32)
    isNearlyEqual = isNearlyEqualLLCore(cmpl, int(compared, kind=int32)&
       , cmpl, precision)
  end function isNearlyEqualInt8Int16

  !---------------------------------------------------------------------
  logical function isNearlyEqualInt16Int16(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    integer(int16), intent(in) :: cmp, compared
    integer,        intent(in) :: precision
    integer(int32) :: cmpl

    cmpl = int(cmp, kind=int32)
    isNearlyEqual = isNearlyEqualLLCore(cmpl, int(compared, kind=int32)&
       , cmpl, precision)
  end function isNearlyEqualInt16Int16

  !---------------------------------------------------------------------
  logical function isNearlyEqualInt16Int8(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    integer(int16), intent(in) :: cmp
    integer(int8),  intent(in) ::      compared
    integer,        intent(in) :: precision
    integer(int32) :: cmpl

    cmpl = int(cmp, kind=int32)
    isNearlyEqual = isNearlyEqualLLCore(cmpl, int(compared, kind=int32)&
       , cmpl, precision)
  end function isNearlyEqualInt16Int8

  !---------------------------------------------------------------------
  logical function isNearlyEqualInt8Int32(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    integer(int8),  intent(in) :: cmp
    integer(int32), intent(in) ::      compared
    integer,        intent(in) :: precision
    integer(int32) :: cmpl

    cmpl = int(cmp, kind=int32)
    isNearlyEqual = isNearlyEqualLLCore(cmpl, compared, cmpl, precision)
  end function isNearlyEqualInt8Int32

  !---------------------------------------------------------------------
  logical function isNearlyEqualInt16Int32(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    integer(int16), intent(in) :: cmp
    integer(int32), intent(in) ::      compared
    integer,        intent(in) :: precision
    integer(int32) :: cmpl

    cmpl = int(cmp, kind=int32)
    isNearlyEqual = isNearlyEqualLLCore(cmpl, compared, cmpl, precision)
  end function isNearlyEqualInt16Int32

  !---------------------------------------------------------------------
  logical function isNearlyEqualInt32Int32(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    integer(int32), intent(in) :: cmp, compared
    integer,        intent(in) :: precision

    isNearlyEqual = isNearlyEqualLLCore(cmp, compared, cmp, precision)
  end function isNearlyEqualInt32Int32

  !---------------------------------------------------------------------
  logical function isNearlyEqualInt32Int16(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    integer(int32), intent(in) :: cmp
    integer(int16), intent(in) ::      compared
    integer,        intent(in) :: precision

    isNearlyEqual = isNearlyEqualLLCore(cmp, int(compared, kind=int32)&
       , cmp, precision)
  end function isNearlyEqualInt32Int16

  !---------------------------------------------------------------------
  logical function isNearlyEqualInt32Int8(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    integer(int32), intent(in) :: cmp
    integer(int8),  intent(in) ::      compared
    integer,        intent(in) :: precision

    isNearlyEqual = isNearlyEqualLLCore(cmp, int(compared, kind=int32)&
       , cmp, precision)
  end function isNearlyEqualInt32Int8

  !---------------------------------------------------------------------
  logical function isNearlyEqualInt8Single(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    integer(int8),  intent(in) :: cmp
    real(single),   intent(in) ::      compared
    integer,        intent(in) :: precision
    real(single) :: cmps

    cmps = real(cmp, kind=single)
    isNearlyEqual = isNearlyEqualRRCore(cmps, compared, cmps, precision)
  end function isNearlyEqualInt8Single

  !---------------------------------------------------------------------
  logical function isNearlyEqualInt16Single(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    integer(int16), intent(in) :: cmp
    real(single),   intent(in) ::      compared
    integer,        intent(in) :: precision
    real(single) :: cmps

    cmps = real(cmp, kind=single)
    isNearlyEqual = isNearlyEqualRRCore(cmps, compared, cmps, precision)
  end function isNearlyEqualInt16Single

  !---------------------------------------------------------------------
  logical function isNearlyEqualInt32Single(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    integer(int32), intent(in) :: cmp
    real(single),   intent(in) ::      compared
    integer,        intent(in) :: precision
    real(single) :: cmps

    cmps = real(cmp, kind=single)
    isNearlyEqual = isNearlyEqualRRCore(cmps, compared, cmps, precision)
  end function isNearlyEqualInt32Single

  !---------------------------------------------------------------------
  logical function isNearlyEqualSingleSingle(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    real(single),   intent(in) :: cmp, compared
    integer,        intent(in) :: precision

    isNearlyEqual = isNearlyEqualRRCore(cmp, compared, cmp, precision)
  end function isNearlyEqualSingleSingle

  !---------------------------------------------------------------------
  logical function isNearlyEqualSingleInt32(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    real(single),   intent(in) :: cmp
    integer(int32), intent(in) ::      compared
    integer,        intent(in) :: precision

    isNearlyEqual = isNearlyEqualRRCore(cmp, real(compared, kind=single)&
       , cmp, precision)
  end function isNearlyEqualSingleInt32

  !---------------------------------------------------------------------
  logical function isNearlyEqualSingleInt16(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    real(single),   intent(in) :: cmp
    integer(int16), intent(in) ::      compared
    integer,        intent(in) :: precision

    isNearlyEqual = isNearlyEqualRRCore(cmp, real(compared, kind=single)&
       , cmp, precision)
  end function isNearlyEqualSingleInt16

  !---------------------------------------------------------------------
  logical function isNearlyEqualSingleInt8(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    real(single),   intent(in) :: cmp
    integer(int8),  intent(in) ::      compared
    integer,        intent(in) :: precision

    isNearlyEqual = isNearlyEqualRRCore(cmp, real(compared, kind=single)&
       , cmp, precision)
  end function isNearlyEqualSingleInt8

  !---------------------------------------------------------------------
  logical function isNearlyEqualInt8Double(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    integer(int8),  intent(in) :: cmp
    real(double),   intent(in) ::      compared
    integer,        intent(in) :: precision
    real(double) :: cmpd

    cmpd = real(cmp, kind=double)
    isNearlyEqual = isNearlyEqualDDCore(cmpd, compared, cmpd, precision)
  end function isNearlyEqualInt8Double

  !---------------------------------------------------------------------
  logical function isNearlyEqualInt16Double(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    integer(int16), intent(in) :: cmp
    real(double),   intent(in) ::      compared
    integer,        intent(in) :: precision
    real(double) :: cmpd

    cmpd = real(cmp, kind=double)
    isNearlyEqual = isNearlyEqualDDCore(cmpd, compared, cmpd, precision)
  end function isNearlyEqualInt16Double

  !---------------------------------------------------------------------
  logical function isNearlyEqualInt32Double(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    integer(int32), intent(in) :: cmp
    real(double),   intent(in) ::      compared
    integer,        intent(in) :: precision
    real(double) :: cmpd

    cmpd = real(cmp, kind=double)
    isNearlyEqual = isNearlyEqualDDCore(cmpd, compared, cmpd, precision)
  end function isNearlyEqualInt32Double

  !---------------------------------------------------------------------
  logical function isNearlyEqualSingleDouble(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    real(single), intent(in) :: cmp
    real(double), intent(in) ::      compared
    integer,      intent(in) :: precision
    real(double) :: cmpd

    cmpd = real(cmp, kind=double)
    isNearlyEqual = isNearlyEqualDDCore(cmpd, compared, cmpd, precision)
  end function isNearlyEqualSingleDouble

  !---------------------------------------------------------------------
  logical function isNearlyEqualDoubleDouble(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    real(double), intent(in) :: cmp, compared
    integer,      intent(in) :: precision

    isNearlyEqual = isNearlyEqualDDCore(cmp, compared, cmp, precision)
  end function isNearlyEqualDoubleDouble

  !---------------------------------------------------------------------
  logical function isNearlyEqualDoubleSingle(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    real(double), intent(in) :: cmp
    real(single), intent(in) ::      compared
    integer,      intent(in) :: precision

    isNearlyEqual = isNearlyEqualDDCore(cmp, real(compared, kind=double)&
       , cmp, precision)
  end function isNearlyEqualDoubleSingle

  !---------------------------------------------------------------------
  logical function isNearlyEqualDoubleInt32(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    real(double),   intent(in) :: cmp
    integer(int32), intent(in) ::      compared
    integer,        intent(in) :: precision

    isNearlyEqual = isNearlyEqualDDCore(cmp, real(compared, kind=double)&
       , cmp, precision)
  end function isNearlyEqualDoubleInt32

  !---------------------------------------------------------------------
  logical function isNearlyEqualDoubleInt16(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    real(double),   intent(in) :: cmp
    integer(int16), intent(in) ::      compared
    integer,        intent(in) :: precision

    isNearlyEqual = isNearlyEqualDDCore(cmp, real(compared, kind=double)&
       , cmp, precision)
  end function isNearlyEqualDoubleInt16

  !---------------------------------------------------------------------
  logical function isNearlyEqualDoubleInt8(cmp, compared, precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
    real(double),   intent(in) :: cmp
    integer(int8),  intent(in) ::      compared
    integer,        intent(in) :: precision

    isNearlyEqual = isNearlyEqualDDCore(cmp, real(compared, kind=double)&
       , cmp, precision)
  end function isNearlyEqualDoubleInt8


  !---------------------------------------------------------------------
  logical function isNearlyEqualDDCore(cmp, compared, cmpnonzero &
       , precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
  ! Algorithm taken from [ruby-list:42510]
  ! 
  ! DESCRIPTION: Compare (all double) inputs of "cmp" and "compared",
  !   based on the digit in "cmpnonzero" (=double, has to be not zero),
  !   for the required order of "precision", and return .true. if
  !   they agree, or .false. otherwise.

    real(double), intent(in) :: cmp, compared, cmpnonzero
    integer,      intent(in) :: precision
    real(double) :: ratio
    integer :: prec

    !-------------------------------------------------------------------
    ! Start:

    if ( precision < 1 ) call error("isNearlyEqualWrongPrecision", '')

    if (cmp == compared) then
      isNearlyEqual = .true.
      return
    else if ((cmp == 0.0) .or. (compared == 0.0)) then
      isNearlyEqual = .false.
      return
    end if

    ratio = abs(cmp/compared)
    if ((ratio < 0.1) .or. (ratio > 10)) then
      isNearlyEqual = .false.
      return
    end if

    prec = precision - floor(log10(abs(cmpnonzero))) - 1
    !if ( nint(cmp * 10.0 ** prec, int32) == nint(compared * 10.0 ** prec, int32) ) then
    if ( abs(cmp*(10.0**prec) - compared*(10.0**prec)) < 1 ) then
      isNearlyEqual = .true.
    else
      isNearlyEqual = .false.
    end if
  end function isNearlyEqualDDCore

  !---------------------------------------------------------------------
  logical function isNearlyEqualRRCore(cmp, compared, cmpnonzero &
       , precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
  ! See isNearlyEqualDDCore() for description.
  ! This is almost exactly the copy of it (but for single-s).

    real(single), intent(in) :: cmp, compared, cmpnonzero
    integer,      intent(in) :: precision
    real(single) :: ratio
    integer :: prec

    !-------------------------------------------------------------------
    ! Start:

    if ( precision < 1 ) call error("isNearlyEqualWrongPrecision", '')

    if (cmp == compared) then
      isNearlyEqual = .true.
      return
    else if ((cmp == 0.0) .or. (compared == 0.0)) then
      isNearlyEqual = .false.
      return
    end if

    ratio = abs(cmp/compared)
    if ((ratio < 0.1) .or. (ratio > 10)) then
      isNearlyEqual = .false.
      return
    end if

    prec = precision - floor(log10(abs(cmpnonzero))) - 1
    !if ( nint(cmp * 10.0 ** prec, int32) == nint(compared * 10.0 ** prec, int32) ) then
    if ( abs(cmp*(10.0**prec) - compared*(10.0**prec)) < 1 ) then
      isNearlyEqual = .true.
    else
      isNearlyEqual = .false.
    end if
  end function isNearlyEqualRRCore

  !---------------------------------------------------------------------
  logical function isNearlyEqualLLCore(cmp, compared, cmpnonzero &
       , precision) result(isNearlyEqual)
  !---------------------------------------------------------------------
  ! See isNearlyEqualDDCore() for description.
  ! This is pretty much the copy of it (but for single-s).

    integer(int32), intent(in) :: cmp, compared, cmpnonzero
    integer,        intent(in) :: precision
    real(double) :: ratio
    integer :: prec

    !-------------------------------------------------------------------
    ! Start:

    if ( precision < 1 ) call error("isNearlyEqualWrongPrecision", '')

    if (cmp == compared) then
      isNearlyEqual = .true.
      return
    else if ((cmp == 0) .or. (compared == 0)) then
      isNearlyEqual = .false.
      return
    end if

    ratio = abs( real(cmp, double)/compared )
    if ((ratio < 0.1) .or. (ratio > 10)) then
      isNearlyEqual = .false.
      return
    end if

    prec = precision - floor(log10(abs(real(cmpnonzero, kind=double)))) - 1
    !if ( nint(cmp * 10.0 ** prec, int32) == nint(compared * 10.0 ** prec, int32) ) then
    if ( abs(cmp*(10.0**prec) - compared*(10.0**prec)) < 1 ) then
      isNearlyEqual = .true.
    else
      isNearlyEqual = .false.
    end if
  end function isNearlyEqualLLCore

end module test_utils
