! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
program testepicaux
  use errorhandling
  use types
  ! use dal		! For INTEGER_NULL
  use cal

  !! Followings are in ssclib
  use epic_aux

  implicit none

  call test_getCcdNodeNum
  call test_inverseCcdNodeNum

contains

  !---------------------------------------------------------------------
  subroutine test_getCcdNodeNum
  !---------------------------------------------------------------------
  ! Tests of getCcdNodeNum() in epic_aux
  !

    implicit none

    character(30)  :: errHead
    integer(int8)  :: ccdNum, nodeNum
    integer(int32) :: ccdNodeNum

    !-------------------------------------------------------------------
    ! Start:

    !----------------------------------------------------------
    errHead = "GetCcdNodeNum001a"
    print *,"=== "//trim(errHead)//": ccdNum=11 (pn)."

    ccdNum = 11
    ccdNodeNum = getCcdNodeNum(ccdNum)

    if ( ccdNodeNum .ne. 11 ) then
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "GetCcdNodeNum002a"
    print *,"=== "//trim(errHead)//": ccdNum=5, nodeNum=1."

    ccdNum = 5
    nodeNum = 1
    ccdNodeNum = getCcdNodeNum(ccdNum, nodeNum)

    if ( ccdNodeNum .ne. 5 ) then
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "GetCcdNodeNum003a"
    print *,"=== "//trim(errHead)//": ccdNum=5, nodeNum=2."

    ccdNum = 5
    nodeNum = 2
    ccdNodeNum = getCcdNodeNum(ccdNum, nodeNum)

    if ( ccdNodeNum .ne. 15 ) then
      call error("err"//trim(errHead), '')
    end if

  end subroutine test_getCcdNodeNum


  !---------------------------------------------------------------------
  subroutine test_inverseCcdNodeNum
  !---------------------------------------------------------------------
  ! Tests of inverseCcdNodeNum() in epic_aux
  !

    implicit none

    character(30)  :: errHead
    integer(int8)  :: ccdNum, nodeNum
    integer(int32) :: ccdNodeNum, instrumentId

    !-------------------------------------------------------------------
    ! Start:

    call CAL_setState(instrument=XMM)

    !----------------------------------------------------------
    errHead = "InverseCcdNodeNum001a"
    print *,"=== "//trim(errHead)//": ccdNodeNum=11 (Default)."

    ccdNodeNum = 11
    instrumentId = -7
    call inverseCcdNodeNum(ccdNodeNum, ccdNum, instrumentId=instrumentId)

    if ( ccdNum .ne. 11 ) then
      print *, ccdNum
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "InverseCcdNodeNum002a"
    print *,"=== "//trim(errHead)//": ccdNodeNum=11 (Default, with nodeNum)."

    ccdNodeNum = 11
    instrumentId = -7
    ! call inverseCcdNodeNum(ccdNodeNum, ccdNum, nodeNum, instrumentId=instrumentId)
    call inverseCcdNodeNum(ccdNodeNum, ccdNum, nodeNum, instrumentId)

    if ( ccdNum .ne. 11  .or.  nodeNum .ne. 0) then
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "InverseCcdNodeNum003a"
    print *,"=== "//trim(errHead)//": ccdNodeNum=11 (EPN, with nodeNum)."

    ccdNodeNum = 11
    instrumentId = EPN
    call inverseCcdNodeNum(ccdNodeNum, ccdNum, nodeNum, instrumentId=instrumentId)

    if ( ccdNum .ne. 11  .or.  nodeNum .ne. 0) then
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "InverseCcdNodeNum004a"
    print *,"=== "//trim(errHead)//": ccdNodeNum=11 (EPN (cal), with nodeNum)."

    ccdNodeNum = 11
    call CAL_setState(instrument=EPN)
    call inverseCcdNodeNum(ccdNodeNum, ccdNum, nodeNum)

    if ( ccdNum .ne. 11  .or.  nodeNum .ne. 0) then
      call error("err"//trim(errHead), '')
    end if
    call CAL_setState(instrument=XMM)

    !----------------------------------------------------------
    errHead = "InverseCcdNodeNum011a"
    print *,"=== "//trim(errHead)//": ccdNodeNum=11 (MOS1)."

    ccdNodeNum = 11
    instrumentId = EMOS1
    call inverseCcdNodeNum(ccdNodeNum, ccdNum, instrumentId=instrumentId)

    if ( ccdNum .ne. 1 ) then
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "InverseCcdNodeNum012a"
    print *,"=== "//trim(errHead)//": ccdNodeNum=11 (MOS2, with nodeNum)."

    ccdNodeNum = 11
    instrumentId = EMOS2
    call inverseCcdNodeNum(ccdNodeNum, ccdNum, nodeNum, instrumentId=instrumentId)

    if ( ccdNum .ne. 1  .or.  nodeNum .ne. 2) then
      print *, ccdNum, nodeNum
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "InverseCcdNodeNum013a"
    print *,"=== "//trim(errHead)//": ccdNodeNum=3 (EPN, with nodeNum)."

    ccdNodeNum = 3
    instrumentId = EMOS2
    call inverseCcdNodeNum(ccdNodeNum, ccdNum, nodeNum, instrumentId=instrumentId)

    if ( ccdNum .ne. 3  .or.  nodeNum .ne. 1) then
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "InverseCcdNodeNum014a"
    print *,"=== "//trim(errHead)//": ccdNodeNum=11 (EMOS1 (cal), with nodeNum)."

    ccdNodeNum = 11
    call CAL_setState(instrument=EMOS1)
    call inverseCcdNodeNum(ccdNodeNum, ccdNum, nodeNum)

    if ( ccdNum .ne. 1  .or.  nodeNum .ne. 2) then
      call error("err"//trim(errHead), '')
    end if
    call CAL_setState(instrument=XMM)

  end subroutine test_inverseCcdNodeNum


end program testepicaux
