! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
program testsourcecutouts1
  use errorhandling
  use types
  use dal		! For INTEGER_NULL
  use cal
  use dss
  use dataFilter	! in selectlib.

  !! Followings are in ssclib
  use array_utils	! for smallestTrueIndex() etc in getBandEdges()
  use cal_aux		! for getCalInfo()
  use dal_aux		! for readColDataToFixed()
  use constants		! for PiOn180
  use coordinate
  use ellipse		! Called from Library source_cutouts
  use source_cutouts
  use test_utils	! for isNearlyEqual()
  use wcs_aux

  implicit none

  character(40), parameter :: gEvlistM1Def &
     = 'outfiles/P0403760401M1S021MIEVLI0000.FIT'
  character(40), parameter :: gSrclistFileNameDef &
     = 'outfiles/P0403760401EPX000OBLSLI0000.FIT'
    != '/data/archive/product/1343_0403760401_001/product/P0403760401M1S021IMAGE_8000.FTZ'

  type(WcsT)                   :: imgWcs, srcWcs
  real(double)                 :: expStartTimeStamp
  type(SpacecraftAttitudeType) :: scAttitude
  real(single)                 :: fracBackground
  type(SrcDataT), allocatable  :: srcData(:) ! typedef in ssclib/source_cutouts


  call prepare_data(imgWcs, srcWcs, expStartTimeStamp, scAttitude &
     , fracBackground, srcData)

  call test_getSourceCutoutsFewArgs

contains

  !---------------------------------------------------------------------
  subroutine test_getSourceCutoutsFewArgs
  !---------------------------------------------------------------------
  ! Tests of getSourceCutoutsFewArgs() (getSourceCutouts() via interface) in source_cutouts

    implicit none

    character(27)  :: errHead
    integer, parameter :: sizeLoc = 3
    type(SrcDataT) :: srcDataTmp(size(srcData)), srcDataLoc(sizeLoc), srcDataBkup(sizeLoc)	! typedef in ssclib/source_cutouts
    real(single) :: bkgFluxPerPixelNow
    real(double) :: s, val
    ! real, parameter :: ThLowerBkgFluxPerPixel = 0.0001	! Copied from source_cutouts.f90

    !-------------------------------------------------------------------
    ! Start:

    srcDataLoc(1) = srcData(1)	! BOX_ID_SRC=1, ID_INST=0, ID_BAND=0
    srcDataLoc(2) = srcData(20)	! BOX_ID_SRC=2, ID_INST=0, ID_BAND=0
    srcDataLoc(3) = srcData(39)	! BOX_ID_SRC=3, ID_INST=0, ID_BAND=0
    srcDataLoc(3)%raDeg  = imgWcs%x%refWorld
    srcDataLoc(3)%decDeg = imgWcs%y%refWorld
    ! srcDataLoc(3) is now located to the image centre, which
    ! should be close to the optical axes.
! call dumpWcs(imgWcs)	! in wcs_aux

    srcDataBkup = srcDataLoc	! For repeated use.

    !----------------------------------------------------------
    errHead = "GetSourceCutoutsFewArgs001a"
    print *,"=== "//errHead//": Warn(allSrcNonPositive)"

    srcDataTmp = srcData
    srcDataTmp%flux = 0
    srcDataTmp%xTan = 0

    call getSourceCutouts(imgWcs, srcWcs, scAttitude, expStartTimeStamp&
       , fracBackground, srcDataTmp)	! issue allSrcNonPositive

    if ( srcDataTmp(1)%xTan .ne. 0 ) then
      call dumpSrcDataT(srcDataTmp(1))
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "GetSourceCutoutsFewArgs002a"
    print *,"=== "//errHead//": Warn(allSrcBkgNonPositive)"

    srcDataTmp = srcData
    srcDataTmp(1)%flux = 0
    srcDataTmp(2:)%bkgFluxPerPixel = 0
    srcDataTmp%xTan = 0

    call getSourceCutouts(imgWcs, srcWcs, scAttitude, expStartTimeStamp&
       , fracBackground, srcDataTmp)	! issue allSrcBkgNonPositive

    if ( (srcDataTmp(1)%xTan .ne. 0) .or. (srcDataTmp(2)%xTan .ne. 0) ) then
      call dumpSrcDataT(srcDataTmp(1), 'Tmp(1)')
      call dumpSrcDataT(srcDataTmp(2), 'Tmp(2)')
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "GetSourceCutoutsFewArgs003a"
    print *,"=== "//errHead//": Check a large-enough value of bkgFluxPerPixel"

    bkgFluxPerPixelNow = 1e-3
    srcDataLoc(2)%bkgFluxPerPixel = bkgFluxPerPixelNow
    call getSourceCutouts(imgWcs, srcWcs, scAttitude, expStartTimeStamp&
       , fracBackground, srcDataLoc)

    if ( srcDataLoc(2)%bkgFluxPerPixel .ne. bkgFluxPerPixelNow ) then
      call dumpSrcDataT(srcDataLoc(2))
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "GetSourceCutoutsFewArgs004a"
    print *,"=== "//errHead//": Check the zero value of bkgFluxPerPixel"

    srcDataLoc = srcDataBkup
    bkgFluxPerPixelNow = 0
    srcDataLoc(2)%bkgFluxPerPixel = bkgFluxPerPixelNow
    srcDataLoc(2)%idNum = 2
    call getSourceCutouts(imgWcs, srcWcs, scAttitude, expStartTimeStamp&
       , fracBackground, srcDataLoc)

    ! if ( srcDataLoc(2)%bkgFluxPerPixel <= bkgFluxPerPixelNow ) then	! Loose constraint
    if ( .not. isNearlyEqual(srcDataLoc(2)%bkgFluxPerPixel &
         , (srcDataLoc(1)%bkgFluxPerPixel + srcDataLoc(3)%bkgFluxPerPixel)/2, 5) ) then	! Strict (or actually exact) constraint.
      print *,'srcDataLoc(1)%bkgFluxPerPixel = ',srcDataLoc(1)%bkgFluxPerPixel
      print *,'srcDataLoc(3)%bkgFluxPerPixel = ',srcDataLoc(3)%bkgFluxPerPixel
      print *,'averageBkgFluxPerPixel =  ', (srcDataLoc(1)%bkgFluxPerPixel + srcDataLoc(3)%bkgFluxPerPixel)/2
      call dumpSrcDataT(srcDataLoc(2))
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "GetSourceCutoutsFewArgs005a"
    print *,"=== "//errHead//": Check a small-enough value of bkgFluxPerPixel"

    srcDataLoc = srcDataBkup
    bkgFluxPerPixelNow = 1e-5
    srcDataLoc(2)%bkgFluxPerPixel = bkgFluxPerPixelNow
    srcDataLoc(2)%idNum = 2
    call getSourceCutouts(imgWcs, srcWcs, scAttitude, expStartTimeStamp&
       , fracBackground, srcDataLoc)

    if ( srcDataLoc(2)%bkgFluxPerPixel <= bkgFluxPerPixelNow ) then	! Loose constraint
      call dumpSrcDataT(srcDataLoc(2))
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "GetSourceCutoutsFewArgs006a"
    print *,"=== "//errHead//": Check ellipse axes ratio"

    srcDataLoc = srcDataBkup
    call getSourceCutouts(imgWcs, srcWcs, scAttitude, expStartTimeStamp&
       , fracBackground, srcDataLoc)

    val = srcDataLoc(3)%region%shortSemiAxis / srcDataLoc(3)%region%longSemiAxis
    if ( (val > 1) .or. (val < 0.8) ) then
      print *,'Short/Long = ', val
      call dumpSrcDataT(srcDataLoc(3))
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "GetSourceCutoutsFewArgs007a"
    print *,"=== "//errHead//": Check xTan, yTan"
      ! Use the above situation from Test 006a.

    if ( (abs(srcDataLoc(3)%xTan) > 1e-4) .or. (abs(srcDataLoc(3)%yTan) > 1e-4) ) then
      call dumpSrcDataT(srcDataLoc(3))
      call error("err"//errHead, '')
    end if

! call dumpSrcDataT(srcDataLoc(3), 'Loc(3)')
! call dumpSrcDataT(srcDataLoc(2), 'Loc(2)')
! call dumpSrcDataT(srcDataLoc(1), 'Loc(1)')

  end subroutine test_getSourceCutoutsFewArgs


  !---------------------------------------------------------------------
  subroutine prepare_data(imgWcs, srcWcs, expStartTimeStamp, scAttitude &
     , fracBackground, srcData)
  !---------------------------------------------------------------------
  ! Taken from region/src/region_mod.f90

    implicit none

    type(WcsT),                   intent(out) :: imgWcs, srcWcs
    real(double),                 intent(out) :: expStartTimeStamp
    type(SpacecraftAttitudeType), intent(out) :: scAttitude
    ! real(single), parameter,      intent(out) :: fracBackground=1.0
    real(single),                 intent(out) :: fracBackground
    type(SrcDataT), allocatable,  intent(out) :: srcData(:) ! typedef in ssclib/source_cutouts

    character(256), parameter ::&
       evlistSetName = trim(gEvlistM1Def),&
       srcListSetTabName = trim(gSrclistFileNameDef)//':SRCLIST',&
       tempSrcSetName = 'outfiles/tmp.ds',&
       ! selectionExpression = 'ID_INST == 2 && ID_BAND == 0',&
       selectionExpression = '',&
       srcRegionSetName = 'outfiles/srcreg.ds',&
       bkgRegionSetName = 'outfiles/bkgreg.ds',&
       bkgKwdSetName = '',&
       bkgImgSetName = ''
    character(20), parameter ::&
       outUnit = 'xy',&	! or 'detxy'
       opMode = 'single',&
       radiusMode = 'contour',&
       srcIdColName = 'BOX_ID_SRC',&	! Only for this test code.
       bkgFluxStyle = 'col' !,&
       ! bkgRateColName = ''
    integer, parameter ::&
       userSrcNum = 4
    real(single), parameter ::&
       userBkgFluxPerPixel = 0.0 !,&
       ! fracEncircEnergy = 0.0,&
       ! fracBackground = 0.0,&
       ! userSrcRadiusArcsec = 0.0
    logical(bool) ::&
       doEllipse = .false. !,&
       ! annulusConcentricAboutFovCtr = .false.,&
       ! retainEllipticalBkgAnnuli = .false.,&
       ! adjustRadius = .false.


    integer(int32) :: instrumentId
    type(SpacecraftAttitudeType) :: instrumAttitude

    type :: localAnnulusType
       real(double) ::&
         centreX,&
         centreY
       real(single) ::&
         innerLongAxis,&
         outerLongAxis
       type(EllipseT) ::& ! typedef in ssclib/ellipse
         ellipse
       character(10) :: unit	! eg 'pixels', 'detxy', 'tanxy'
       logical(bool) :: isAnnulus	! FALSE means the inner area is of zero size.
       ! This should override any non-zero value in innerLongAxis.
    end type localAnnulusType

  !   type :: localRegListType
  !      type(ims_RegionT), pointer :: regions(:)
  !   end type localRegListType
  ! 
  ! ! --- from region/ims_ascregion_mod.f90
  ! integer, parameter :: mySHAPE_LEN = 16
  ! integer, parameter :: myMAX_POLY = 4  !!!!!50
  ! 
  ! type :: ims_RegionT
  !   character(mySHAPE_LEN) ::&
  !     shape
  !   real(single) ::&
  !     x(myMAX_POLY),&
  !     y(myMAX_POLY),&
  !     r(myMAX_POLY),&
  !     a(myMAX_POLY)
  !   integer(int32) ::&
  !     component
  ! end type ims_RegionT
  ! 
  ! type :: ims_RegionExtnT
  !   character(8) ::&
  !     extName
  !   character(20) ::&
  !     xColName,&
  !     yColName
  !   logical(bool) ::&
  !     withWcs
  !   type(WcsT) ::&
  !     wcs
  !   type(ims_RegionT), pointer ::&
  !     regions(:)
  ! end type ims_RegionExtnT
  ! ! --- up to here (from region/ims_ascregion_mod.f90)

    type(DataSetT) :: set, sourceSet
    type(TableT) :: tab, sourceTab
    character(255) :: srcListSetName, srcListTabName
    integer(int32) :: numRows, numFilteredRows
    integer :: srcNum &
       , trialIdInst, trialIdBand, xi, yi, xiSize, yiSize, i
    integer, parameter :: maxSrcNum = 9999
    integer(int32), allocatable :: srcIdBand(:), srcIdInst(:)
    real(double) ::&
      bandLo(3,0:9), bandHi(3,0:9), meanEeV,& ! userSrcRadiusTan,&
      ! smallestPixelDim,&
      pixelSizeXWorld, pixelSizeYWorld
    logical(bool) :: infoFoundThisBand(3,0:9)
    real(double), allocatable :: tanX(:), tanY(:)
    ! real(single), allocatable :: srcShrinkage(:)
    ! type(localAnnulusType), pointer :: bkgAnnulus(:)
    ! type(ims_RegionExtnT), allocatable :: regionExtnData(:)
    ! type(ims_RegionT) :: srcRegion
    ! type(ims_RegionT), pointer :: bkgRegions(:)
    ! type(localRegListType), pointer :: bkgRegList(:)
    ! type(ims_RegionT), allocatable :: srcRegions(:)
    ! character(4) :: srcIdStr
    character(20) :: xColName, yColName
    real(single), pointer :: image(:,:)
logical, parameter :: localTest = .true.

! real(single), allocatable :: testSingle(:)
! real(double), allocatable :: testDouble(:)
! integer(int16), allocatable :: testInt16(:)


    !-------------------------------------------------------------------
    ! Start

    fracBackground=1.0	! Fixed

    ! call readParameters

    select case(outUnit)
    case('xy')
      xColName = 'X'
      yColName = 'Y'
    
    case('detxy')
      xColName = 'DETX'
      yColName = 'DETY'

    case default
      call error('badOutUnit',"The value of the --outunit parameter is invalid.")
    end select

    !...................................................................
    ! Set up the cal and get wcs info if appropriate:

    call message(AppMsg, verbose, "Obtaining essential information from &
      &the event list.")

    set = dataSet(evlistSetName, READ)

    if (outUnit == 'xy') then
      tab = table(set, 'EVENTS')
      call readWcs(tab, 'X', 'Y', imgWcs) ! in ssclib/wcs_aux
    
      pixelSizeXWorld = sqrt(imgWcs%x%matrixOnAxis * imgWcs%x%matrixOnAxis&
       + imgWcs%x%matrixOffAxis * imgWcs%x%matrixOffAxis)
      pixelSizeYWorld = sqrt(imgWcs%y%matrixOnAxis * imgWcs%y%matrixOnAxis&
       + imgWcs%y%matrixOffAxis * imgWcs%y%matrixOffAxis)
    !*** this is only valid if there is no skew.
    end if

    call getCalInfo(set, instrumentId, expStartTimeStamp, scAttitude&
      , instrumAttitude) ! in ssclib/cal_aux. Sets CAL instrument state.

    if (instrumentId .ne. EMOS1&
    .and. instrumentId .ne. EMOS2&
    .and. instrumentId .ne. EPN) then
call error('badInstrum', 'Instrument '//trim(instId(instrumentId))//&
&' not supported.')
    end if

    call release(set)

    !...................................................................
    ! Filter and load the source list:
!*** exactly the same as in imageutils/srcmask.

    call message(AppMsg, Verbose, 'Filtering source list.')

    call splitSetTabName(srcListSetTabName, srcListSetName, srcListTabName)

    ! in ssclib/dal_aux
    sourceSet = clone(srcListSetName, tempSrcSetName, MODIFY)
    sourceTab = table(sourceSet, srcListTabName, MODIFY)

    numRows = numberOfRows(sourceTab)
    write(errStr, '(a, i8)') "Original NumberOfRows = ", numRows
    call message(AppMsg, Verbose)
    if (trim(selectionExpression) == '') then
      numFilteredRows = numRows
      call message(AppMsg, Verbose, 'No filtering is applied.')
    else
      call initFilter(selectionExpression)
      numFilteredRows = doFilter(sourceTab)
      write(errStr, '(a, i8)') "Filtered NumberOfRows = ", numFilteredRows
      call message(AppMsg, Verbose)
    end if

    if (numFilteredRows <= 0)&
      call warning('noSourcesSelected'&
        , 'No sources passed the specified filtering expression.')

    if (userSrcNum > numFilteredRows) then
      write(errStr, '(a, i5, a)') "There are fewer rows (= "&
         , numFilteredRows, ") in the filtered source list than --usersrcid."
call error('tooFewRows')
    end if
!if (userSrcNum > numFilteredRows) call error('tooFewRows',"There &
!&are fewer rows in the filtered source list than --usersrcid.")
!!if (userSrcNum > maxSrcNum) call error('userSrcNumTooHigh',"The value &
!!&of the --usersrcid parameter exceeds the maximum allowed.")

    if (numFilteredRows > maxSrcNum) then
      call warning('tooManyRows',"The number of rows in the filtered source &
        &list exceeds the maximum value of --usersrcid.")
    end if

    ! if (outUnit == 'xy' .and. radiusMode .eq. 'contour') then
      call readWcsSourceList(sourceSet, srcWcs)	! Test of contour.
    ! !*** they should be stored as proper column keywords, but until Potsdam change their code there's not much I can do.
    ! end if

    allocate(srcData(numFilteredRows))
! allocate(testSingle(numFilteredRows), testDouble(numFilteredRows)&
!   , testInt16(numFilteredRows))
    call initializeSrcDataT(srcData) ! in ssclib/source_cutouts
    srcData%isValid = .true. ! default

    allocate(srcIdBand(numFilteredRows), srcIdInst(numFilteredRows))

    call readColDataToFixed(sourceTab, 'RA',         srcData%raDeg)
    call readColDataToFixed(sourceTab, 'DEC',        srcData%decDeg)
    if (radiusMode .eq. 'contour') then

      call readColDataToFixed(sourceTab, 'SCTS',       srcData%flux)
      call readColDataToFixed(sourceTab, 'BG_MAP',     srcData%bkgFluxPerPixel)

      if (bkgFluxStyle .eq. 'image') then
        call readColDataToFixed(sourceTab, 'X_IMA', srcData%xPixel)
        call readColDataToFixed(sourceTab, 'Y_IMA', srcData%yPixel)
      end if

    else
      srcData%flux = 0.0
      srcData%bkgFluxPerPixel = 0.0
    end if

    if (radiusMode .ne. 'userfixed') then
      call readColDataToFixed(sourceTab, 'ID_INST',    srcIdInst)
      call readColDataToFixed(sourceTab, 'ID_BAND',    srcIdBand)
    end if

    if (opMode .eq. 'batch') then
      if (trim(srcIdColName) == '' .or. hasAttribute(sourceTab, trim(srcIdColName))) then
        call readColDataToFixed(sourceTab, srcIdColName, srcData%idNum)	! in ssclib/dal_aux.
      else
        call error('notExistsSrcidcol',"--srcidcol of ("//trim(srcIdColName)&
           //") does not exist as an attribute in ("&
           //trim(srcListSetTabName)//").")
        !! The row number in the original FITS table may not agree with
        ! the current row number because it is now dealing with 
        ! the filtered table, based on the given selExpression.
      end if
    end if

    call getBandEdges(sourceSet, bandLo, bandHi, infoFoundThisBand)

    call release(sourceSet)

    if (opMode == 'batch') then
      do srcNum = 1, numFilteredRows
        if (count(srcData%idNum == srcData(srcNum)%idNum) > 1)&
call error('nonUniqueSrcId',"The &
&source ID number is not unique in the filtered source set.")
      end do

    else if (opMode == 'single') then
      srcData(userSrcNum)%idNum = userSrcNum
    end if

    if (opMode .ne. 'single' .and. numFilteredRows > 0) then
      ! The latter condition is to guarantee srcIdBand(1) exists.
      if (count(srcIdBand == srcIdBand(1)) .ne. numFilteredRows) then
        call warning('nonUniqueIdBand',"More than one value of ID_BAND was &
        &detected. This implies that several concentric regions may be &
        &produced for a single source. If you don't want this to happen you &
        &should filter the source list (the task can do this for you via its &
        &--expression parameter) so that only one value of ID_BAND remains.")
      end if
    end if

    if (radiusMode .ne. 'userfixed') then
      !.................................................................
      ! Calculate source energies:

      call message(AppMsg, verbose, "Calculating a representative energy &
        &for each source.")

      if (numFilteredRows > 0) then	! to guarantee srcIdBand(1) exists.
if (srcIdBand(1) < 0 .or. srcIdBand(1) > 9) call error('badIdBand',"The &
&value of the ID_BAND column of the source list must be between 0 and 9 &
&inclusive.")
      end if

      srcData%EeV = 500.0 ! default.
      do trialIdInst = 1, 3
        do trialIdBand = 0, 9
          if (count(srcIdInst == trialIdInst&
          .and. srcIdBand == trialIdBand)<=0)&
        cycle

          if (count(srcIdInst == trialIdInst.and.srcIdBand == trialIdBand) > 0&
          .and. .not. infoFoundThisBand(trialIdInst, trialIdBand)) then
            call warning('noBandEnergy',"No energy bounds were found for &
              &sources of this value of ID_INST and ID_BAND. The default value &
              &of mean source energy has been used instead.")
!*** put in mumbers.
        cycle
          end if

if (bandLo(trialIdInst, trialIdBand) >= bandHi(trialIdInst, trialIdBand))&
call error('badEnergyBandEdges',"The 'lower' energy is higher than the &
&'higher' energy for this band!")

          meanEeV = (bandLo(trialIdInst, trialIdBand)&
            + bandHi(trialIdInst, trialIdBand)) / 2.0

          where(srcIdInst == trialIdInst .and. srcIdBand == trialIdBand)&
            srcData%EeV = meanEeV
        end do
      end do
    end if

    !...................................................................
    ! Get the positions of the source centres in the appropriate coord
    ! system:

    call message(AppMsg, verbose, "Calculating the location of each source &
      &in the specified coordinate system.")

    ! There are combinations of command-line parameters for which it is
    ! unnecessary to calculate either DETXY or TELCOORD positions of the
    ! sources. However rather than set up a complicated 'if' statement
    ! which might then become incorrect if facilities were added or
    ! changed at a later date, I have decided that it would be simpler
    ! just to calculate both DETXY or TELCOORD positions under all
    ! circumstances. The time penalty involved is insignificant.

    call raDecToInst(srcData%raDeg*PiOn180, srcData%decDeg*PiOn180, scAttitude&
     , expStartTimeStamp, srcData%xDet, srcData%yDet, srcData%thetaArcsec&
     , srcData%phi) ! in ssclib/coordinate

    select case(outUnit)
    case('xy')
      if (radiusMode == 'userfixed' .or. radiusMode == 'enfrac') then

        allocate(tanX(numFilteredRows), tanY(numFilteredRows))
        call raDecToTan(srcData%raDeg*PiOn180, srcData%decDeg*PiOn180&
         , imgWcs%x%refWorld*PiOn180, imgWcs%y%refWorld*PiOn180&
         , tanX, tanY) ! in ssclib/coordinate

        call tanToPixels(imgWcs, tanX, tanY, srcData%xPixel, srcData%yPixel)
        ! in ssclib/wcs_aux

        deallocate(tanX, tanY)
      end if
      ! Note that if radiusMode == 'contour' then source centre xy positions
      ! are obtained later as a by-product of calculating the hole locii.

    case('detxy')
!      if (radiusMode == 'userfixed') then
!        call raDecToInst(srcData%raDeg, srcData%decDeg, scAttitude&
!          , expStartTimeStamp, srcData%xDet, srcData%yDet)
!         ! in ssclib/coordinate
!      end if
!      ! Note that if radiusMode == 'contour' then source centre detxy positions
!      ! are obtained later as a by-product of calculating the hole locii;
!      ! whereas if radiusMode == 'enfrac', detxy positions have already been
!      ! obtained.

    case default
call error('badOutUnit',"The value of the --outunit parameter is invalid.")
    end select


    !...................................................................
    ! Get the edges of the source holes:

    call message(AppMsg, verbose, "Calculating the source region locii.")

    select case(radiusMode)
    case('contour')
      select case(bkgFluxStyle)
      case('col') ! stick with what we have
      case('kwd')
!*** must divide by exposure?? Don't allow this mode yet.
call error('notSupported', "This mode is not yet supported.")
        set = dataSet(bkgKwdSetName, READ)
        srcData%bkgFluxPerPixel = real32Attribute(set, 'BKG_RATE')
        call release(set)

      case('image')
!call error('notSupported', "This mode is not yet supported.")
        call readArrayData(bkgImgSetName, image) ! ssclib/dal_aux
        xiSize = size(image, 1)
        yiSize = size(image, 2)

        do i = 1, numFilteredRows
          xi = nint(srcData(i)%xPixel)
          yi = nint(srcData(i)%yPixel)
          if (xi < 1 .or. xi > xiSize .or. yi < 1 .or. yi > yiSize) then
            srcData(i)%isValid = .false.
          else
            srcData(i)%bkgFluxPerPixel = image(xi, yi)
            if (image(xi, yi) <= 0.0) srcData(i)%isValid = .false.
          end if
        end do

        deallocate(image)

      case('user')
        srcData%bkgFluxPerPixel = userBkgFluxPerPixel
      case default
call error('badBkgFluxStyle', "The value of --bkgratestyle not recognized.")
      end select

! 
! !*** temporarily:
! if (outUnit == 'detxy') call error('notSupported',"--radiusstyle=contour, &
! &--outunit=detxy is not yet supported.")
! !** Note: if (outUnit == 'detxy') then srcWcs a couple of lines below is undefined.
! 
! print *,'here00010: ',userSrcNum
! if (localTest) &
! call dumpSrcDataT(srcData(userSrcNum),'srcData(4) --before')
!       call getSourceCutouts(imgWcs, srcWcs, scAttitude, expStartTimeStamp&
!         , fracBackground, srcData)
! !******** should modify this so it tests for isValid.
!       ! in ssclib/source_cutouts.
!       !
!       ! Note! This returns both xy and detxy coords of the source positions
!       ! in the srcData structure.
! if (localTest) &
! call dumpSrcDataT(srcData(userSrcNum),'srcData(4) --after')
! 
! !!!      if (outUnit == 'detxy') then
! !*** transfer src region to detxy.
! !!!      end if
! 
!       if (doEllipse) then
!         srcData%region%isCircle = .false.
!       else
!         srcData%region%isCircle = .true.
!         srcData%region%shortSemiAxis = srcData%region%longSemiAxis
!       end if
! 

    case default
      ! call error('badRadiusStyle', "The value of --radiusstyle not recognized.")
      call error('badRadiusStyle', "The value of --radiusstyle is not appropriate here.")
    end select

  end subroutine prepare_data


  ! Extracted from region/src/region_mod.f90
  !---------------------------------------------------------------------
  subroutine getBandEdges(set, bandLo, bandHi, infoFoundThisBand)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: set
    real(double), intent(out) :: bandLo(3,0:9), bandHi(3,0:9)
    logical(bool), intent(out) :: infoFoundThisBand(3,0:9)

    character(2) :: instChars(3)
    integer :: idInst, idBand
    character(8) :: loKwdName, hiKwdName

    !-------------------------------------------------------------------
    ! Start

    instChars = (/'PN', 'M1', 'M2'/)

    infoFoundThisBand = .false. ! default
    bandLo = 0.0 ! default
    bandHi = 0.0 ! default
    do idInst = 1, 3
      do idBand = 1, 9
        write(loKwdName, '(a3, i1, a4)') instChars(idInst)//'_', idBand, '_ELO'
        write(hiKwdName, '(a3, i1, a4)') instChars(idInst)//'_', idBand, '_EHI'
        if (hasAttribute(set, loKwdName)&
        .and. hasAttribute(set, hiKwdName)) then
          infoFoundThisBand(idInst,idBand) = .true.
          bandLo(idInst,idBand) = real64Attribute(set, loKwdName)
          bandHi(idInst,idBand) = real64Attribute(set, hiKwdName)
        end if
      end do

      if (any(infoFoundThisBand(idInst,1:9))) then 
        infoFoundThisBand(idInst,0) = .true.
        bandLo(idInst,0) = bandLo(idInst&
          , smallestTrueIndex(infoFoundThisBand(idInst,1:9)))
        bandHi(idInst,0) = bandHi(idInst&
          ,  largestTrueIndex(infoFoundThisBand(idInst,1:9)))
        ! These calls are in ssclib/array_utils.
      end if
    end do

  end subroutine getBandEdges

end program testsourcecutouts1
